/*
  kitchenclock.cpp - Kitschy Nixie Kitchen Clock

  Copyright (c) 2017 Joe Croft <croftj@gmail.com>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA

  $Id: //depot/kitchenclock/code/kitchenclock.cpp#8 $
*/

# include <avr/pgmspace.h>
# include <ctype.h>
# include <math.h>
# include <stdio.h>
# include <stdlib.h>

# include "clock.h"
# include "datetime.h"
# include "esquare.h"
# include "Hands.h"
# include "switch.h"
# include "core/HardwareSerial.h"
# include "Arduino.h"
# include "IOPort.h"

#define  V_INIT      0
#define  V_INTERVAL  5000
#define  V_PAUSE     2000
#define  V_SAMPLES   40
#define  V_START     125

#define  SPIN_LOOPS  5

#define  WAIT_TEST   20

#define  MIN_ADVANCE_SLOW     +1
#define  MIN_ADVANCE_SLOW_BKW -1
#define  MIN_ADVANCE_FAST     +10
#define  MIN_ADVANCE_FAST_BKW -10
#define  HOUR_ADVANCE_FWD     +60
#define  HOUR_ADVANCE_BKW     -60

#define CLOCK_UPDATE_INTERVAL 300

extern OutputPort gen_60hz;
extern OutputPort dbg2;
extern OutputPort dbg3;
extern OutputPort dbg4;

InputPort AdjustHands;
InputPort DbgIO;

char p_buf[128];
char ibuf[32];
uint16_t pause_dt[Clock::NUM_ELEMENTS];

bool dbg = true;

enum 
{
   CLOCK_RUNNING,
   CLOCK_FORWARD,
   CLOCK_SW1_ON,
   CLOCK_SW2_ON,
   CLOCK_SLOW_FORWARD,
   CLOCK_FAST_FORWARD,
   CLOCK_PAUSED
};

void AdvanceMinutes(int8_t increment)
{
   int8_t hr_inc = increment / 60;
   int8_t min_inc = increment % 60;
   if (dbg) Serial << __PRETTY_FUNCTION__ << strcpy_P(p_buf, PSTR("increment = ")) << (int)increment;
   if (dbg) Serial << strcpy_P(p_buf, PSTR(", hr_inc = ")) << (int)hr_inc;
   if (dbg) Serial << strcpy_P(p_buf, PSTR(", min_inc = ")) << (int)min_inc << endl;

   uint16_t dt[Clock::NUM_ELEMENTS];
   currentDate.getTime(dt);
   uint8_t secs  = dt[Clock::Second];
   uint8_t mins  = dt[Clock::Minute] + min_inc;
   int8_t hrs   = dt[Clock::Hour] + hr_inc;
   if ( hrs < 0 ) 
   {
      hrs = 23;
   }
   if ( mins >= 60 ) 
   {
      mins -= 60;
      if ( ++hrs >= 24 ) 
      {
         hrs = 0;
      }
   }

   dt[Clock::Minute] = mins;
   dt[Clock::Hour] = hrs;
   currentDate.setTime(dt);

   if (dbg) Serial << __PRETTY_FUNCTION__ << "New Time: " << (uint16_t)dt[Clock::Hour]; 
   if (dbg) Serial << ":" << (uint16_t)dt[Clock::Minute]; 
   if (dbg) Serial << ":" << (uint16_t)dt[Clock::Second]; 
   if (dbg) Serial << " " << (uint16_t)dt[Clock::Month]; 
   if (dbg) Serial << "/" << (uint16_t)dt[Clock::Day]; 
   if (dbg) Serial << "/" << (uint16_t)dt[Clock::Year] << endl; 
   
   uint8_t hrs_hand = hrs % 12;
   if ( hrs_hand == 0 ) 
   {
      hrs_hand = 12;
   }
   uint8_t min_hand = mins / 5;
   if ( min_hand == 0 ) 
   {
      min_hand = 12;
   }
   uint8_t sec_hand = secs % 12 + 1;
   uint8_t sec_hands = mins % 5 + 1;

   Hands::instance()->setSecondHand(sec_hand, sec_hands, false);
   if ( min_hand != Hands::instance()->minutes() ) 
   {
      Hands::instance()->setMinuteHand(min_hand, false);
   }
   else 
   {
      Hands::instance()->setMinuteHand(min_hand, false);
   }

   Hands::instance()->setHourHand(hrs_hand, false);
}

bool refreshMinute()
{
   uint8_t start = Hands::instance()->minutes();
   if ( start == 0 ) 
   {
      start = 1;
   }
   dbg4.on();
//   Serial << strcpy_P(p_buf, PSTR("refreshMinute() Clearing hands: start = ")) << (int) start;
   Hands::instance()->setHand(start, 0, false);

   for (uint8_t x = 2 * 12 - 2; x > 0; x--) 
   {
      if ( --start == 0 )
      {
         start = 12;
      }
      if (dbg) Serial << strcpy_P(p_buf, PSTR(", next = ")) << (int) start;

      Hands::instance()->forceHand(start, 100);
      Clock::delay(125);
   }
   if (dbg) Serial << endl;
   dbg4.off();
//   Hands::instance()->setHand(start, 100, false);
}
   
int main(void)
{
   /***********************************************/
   /*   Initialize  the  serial port and show we  */
   /*   had a reset!                              */
   /***********************************************/
   init();
   Serial.begin(115200); //hello

   Serial << endl;
   Serial << endl;
   Serial << strcpy_P(p_buf, PSTR("/************                                    ************/")) << endl;
   Serial << strcpy_P(p_buf, PSTR("/************           R  E  S  E  T            ************/")) << endl;
   Serial << strcpy_P(p_buf, PSTR("/************                                    ************/")) << endl;
   Serial << strcpy_P(p_buf, PSTR("/*** $Id: //depot/kitchenclock/code/kitchenclock.cpp#8 $  ***/")) << endl;
   Serial << strcpy_P(p_buf, PSTR("/************          $Change: 170 $            ************/")) << endl;
   eeprom.read();
   ESquareSpace_t* settings = eeprom.settings();

   DbgIO.SetPort(IOPort::PORT_B, _BV(0));
   DbgIO.on();

   dbg = DbgIO.read() != 0;

   AdjustHands.SetPort(IOPort::PORT_B, _BV(1));
   AdjustHands.on();

   Hands::instance()->setESquare(eeprom.settings());
   eeprom.read();

   switches.setAutoRepeat(true);
   switches.setAutoRepeat(SWITCH3, false);
   switches.setAutoRepeat(SWITCH4, false);

   /***********************************************/
   /*   Have  'clock'  create  the hands and use  */
   /*   them.                                     */
   /***********************************************/
   currentDate.init();
   currentDate.setHands();
   uint16_t dt[7];
   currentDate.readClock(dt);
   if ( dt[Clock::Hour] > 23 )
   {
      Serial << __PRETTY_FUNCTION__ << strcpy_P(p_buf, PSTR(" Clock had errant time, resetting date and time")) << endl; 
      dt[Clock::Second] = 0;
      dt[Clock::Minute] = 0;
      dt[Clock::Hour]   = 0;  
      dt[Clock::Day]    = 1;
      dt[Clock::Month]  = 03;
      dt[Clock::Year]   = 2017;
   }
   Serial << endl; 
   Serial << __PRETTY_FUNCTION__ << " Start: " << (uint16_t)dt[Clock::Hour]; 
   Serial << ":" << (uint16_t)dt[Clock::Minute]; 
   Serial << ":" << (uint16_t)dt[Clock::Second]; 
   Serial << " " << (uint16_t)dt[Clock::Month]; 
   Serial << "/" << (uint16_t)dt[Clock::Day]; 
   Serial << "/" << (uint16_t)dt[Clock::Year] << endl; 
   Serial << endl; 
         
   for (uint8_t hand = 0; hand < 12; hand++) 
   {                    
      Serial << __PRETTY_FUNCTION__ << " reseting hand "<< (int)hand << endl; 
      Hands::instance()->forceHand(hand, 0x00);
      Hands::instance()->writeHand(hand, 0x00);
      delay(100);
   }
   delay(250);

   Serial << __PRETTY_FUNCTION__ << " done!" << endl; 
   Clock::delay(500);
   Serial << __PRETTY_FUNCTION__ << " continuing" << endl; 

   uint8_t secs = dt[Clock::Second];
   uint8_t mins = dt[Clock::Minute];
   uint8_t hrs = dt[Clock::Hour];
   uint8_t hrs_hand = hrs % 12;
   if ( hrs_hand == 0 ) 
   {
      hrs_hand = 12;
   }

   uint8_t min_hand = mins / 5;
   if ( min_hand == 0 ) 
   {
      min_hand = 12;
   }
   uint8_t sec_hand = secs % 12 + 1;
   uint8_t sec_hands = mins % 5 + 1;
   Serial << strcpy_P(p_buf, PSTR("Hours: ")) << (int)hrs;
   Serial << strcpy_P(p_buf, PSTR(", minutes: ")) << (int)mins;
   Serial << strcpy_P(p_buf, PSTR(", seconds: ")) << (int)secs;
   Serial << strcpy_P(p_buf, PSTR(", hrs hand: ")) << (int)hrs_hand;
   Serial << strcpy_P(p_buf, PSTR(", min hand: ")) << (int)min_hand;
   Serial << strcpy_P(p_buf, PSTR(", sec hand: ")) << (int)sec_hand;
   Serial << strcpy_P(p_buf, PSTR(", sec hands: ")) << (int)sec_hands;
   Serial << endl;

   mins = 1;

   char *line_p;
   char line_buf[120];
   uint8_t clock_state = CLOCK_RUNNING;

   while (1)
   {
      uint16_t dt[7];
      gen_60hz.on();

      /***********************************************/
      /*   Check  if  we are in normal run mode, if  */
      /*   so  (jumper  not in place), then let the  */
      /*   clock run normally.                       */
      /***********************************************/
      if ( 1 && AdjustHands.read() != 0 ) 
      {
         if ( currentDate.getInterval(Clock::CLOCK_UPDATE_TIMER) > CLOCK_UPDATE_INTERVAL )
         {
            dbg2.on();
            if (dbg) Serial << strcpy_P(p_buf, PSTR("Synchronizing SW clock with the RTC")) << endl;
            currentDate.resetInterval(Clock::CLOCK_UPDATE_TIMER);
            currentDate.updateFromRTC();
            dbg2.off();
         }
         int8_t ch = switches.check();
         if ( ch >= 0 ) 
         {
            Serial << strcpy_P(p_buf, PSTR("ch: ")) << (int)ch << endl;
            Serial << strcpy_P(p_buf, PSTR("clock_state: ")) << (int)clock_state << endl;
         }
         
         switch(clock_state)
         {
            case CLOCK_FAST_FORWARD:
            {
               if ( ch == SW2_SHORT || ch == SW2_ON ) 
               {
                  currentDate.readClock(pause_dt);
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_PAUSED")) << endl;
                  clock_state = CLOCK_PAUSED;
               }
               else if ( ch == SW1_SHORT ) 
               {
                  Serial << strcpy_P(p_buf, PSTR("Advancing clock by: ")) << (int)MIN_ADVANCE_FAST << endl;
                  AdvanceMinutes(MIN_ADVANCE_FAST);
               }
               else if ( ch == SW1_OFF )
               {
                  currentDate.getTime(pause_dt);
                  pause_dt[Clock::Second] = 0;
                  currentDate.setTime(pause_dt);

                  time_t now = currentDate.now();
                  Serial << "Setting HW Clock to: " << DateTime::formatTime(now) << endl;
                  currentDate.writeClock(pause_dt);
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_RUNNING")) << endl;
                  clock_state = CLOCK_RUNNING;
               }
            }
            break;

            case CLOCK_SLOW_FORWARD:
            {
               if ( ch == SW1_SHORT || ch == SW1_ON ) 
               {
                  currentDate.readClock(pause_dt);
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_PAUSED")) << endl;
                  clock_state = CLOCK_PAUSED;
               }
               else if ( ch == SW2_SHORT ) 
               {
                  Serial << strcpy_P(p_buf, PSTR("Advancing clock by: ")) << (int)MIN_ADVANCE_SLOW << endl;
                  AdvanceMinutes(MIN_ADVANCE_SLOW);
               }
               else if ( ch == SW2_OFF )
               {
                  currentDate.getTime(pause_dt);
                  pause_dt[Clock::Second] = 0;
                  currentDate.setTime(pause_dt);

                  time_t now = currentDate.now();
                  Serial << "Setting HW Clock to: " << DateTime::formatTime(now) << endl;
                  currentDate.writeClock(pause_dt);
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_RUNNING")) << endl;
                  clock_state = CLOCK_RUNNING;
               }
            }
            break;

            case CLOCK_PAUSED:
            {
               if ( ch == SW1_OFF || ch == SW2_OFF ) 
               {
                  pause_dt[Clock::Second] = 0;
                  currentDate.setTime(pause_dt);
                  currentDate.writeClock(pause_dt);
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_RUNNING")) << endl;
                  clock_state = CLOCK_RUNNING;
                  refreshMinute();
               }
            }
            break;

            case CLOCK_SW1_ON:
            {
               if ( ch == SW2_ON ) 
               {
                  currentDate.readClock(pause_dt);
                  clock_state = CLOCK_PAUSED;
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_PAUSED")) << endl;
               }
               else if ( ch == SW1_SHORT )
               {
                  AdvanceMinutes(MIN_ADVANCE_FAST);
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_FAST_FORWARD")) << endl;
                  clock_state = CLOCK_FAST_FORWARD;
               }
               else if ( ch >= 0 )
               {
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_RUNNING")) << endl;
                  clock_state = CLOCK_RUNNING;
               }
            }
            break;

            case CLOCK_SW2_ON:
            {
               if ( ch == SW1_ON ) 
               {
                  currentDate.readClock(pause_dt);
                  clock_state = CLOCK_PAUSED;
               }
               else if ( ch == SW2_SHORT )
               {
                  AdvanceMinutes(MIN_ADVANCE_SLOW);
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_SLOW_FORWARD")) << endl;
                  clock_state = CLOCK_SLOW_FORWARD;
               }
               else if ( ch >= 0 )
               {
                  Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_RUNNING")) << endl;
                  clock_state = CLOCK_RUNNING;
               }
            }
            break;
         }

         if ( clock_state == CLOCK_RUNNING ) 
         {
            if ( ch == SW1_ON ) 
            {
               dbg2.on();
               Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_SW1_ON")) << endl;
               currentDate.updateFromRTC();
               clock_state = CLOCK_SW1_ON;
               dbg2.off();
            }
            else if ( ch == SW2_ON ) 
            {
               dbg2.on();
               Serial << strcpy_P(p_buf, PSTR("Moving to CLOCK_SW2_ON")) << endl;
               currentDate.updateFromRTC();
               clock_state = CLOCK_SW2_ON;
               dbg2.off();
            }
            else if ( ch == SW3_ON ) 
            {
               dbg3.on();
               Serial << strcpy_P(p_buf, PSTR("CLOCK_SW3_ON")) << endl;
               AdvanceMinutes(HOUR_ADVANCE_FWD);
               uint16_t dt[7];
               currentDate.getTime(dt);
               time_t now = currentDate.now();
               Serial << "Setting HW Clock to: " << DateTime::formatTime(now) << endl;
               currentDate.writeClock(dt);
               dbg3.off();
            }
            else if ( ch == SW4_ON ) 
            {
               dbg3.on();
               Serial << strcpy_P(p_buf, PSTR("CLOCK_SW4_ON")) << endl;
               AdvanceMinutes(HOUR_ADVANCE_BKW);
               uint16_t dt[7];
               currentDate.getTime(dt);
               time_t now = currentDate.now();
               Serial << "Setting HW Clock to: " << DateTime::formatTime(now) << endl;
               currentDate.writeClock(dt);
               dbg3.off();
            }
            else 
            {
               if ( currentDate.getInterval(Clock::ONE_SECOND_TICK) > 0 )
               {
                  currentDate.resetInterval(Clock::ONE_SECOND_TICK);
                  uint16_t dt[7];
                  currentDate.readClock(dt);

                  uint8_t secs = dt[Clock::Second];
                  uint8_t mins = dt[Clock::Minute];
                  uint8_t hrs = (dt[Clock::Hour]);
                  uint8_t hrs_hand = hrs % 12;
                  uint8_t min_hand = mins / 5;
                  uint8_t sec_hand = secs % 12;
                  uint8_t sec_hands = mins % 5 + 1;
                  if (dbg) Serial << strcpy_P(p_buf, PSTR("Hours: ")) << (int)hrs;
                  if (dbg) Serial << strcpy_P(p_buf, PSTR(", minutes: ")) << (int)mins;
                  if (dbg) Serial << strcpy_P(p_buf, PSTR(", seconds: ")) << (int)secs;
                  if (dbg) Serial << strcpy_P(p_buf, PSTR(", hrs hand: ")) << (int)hrs_hand;
                  if (dbg) Serial << strcpy_P(p_buf, PSTR(", min hand: ")) << (int)min_hand;
                  if (dbg) Serial << strcpy_P(p_buf, PSTR(", sec hand: ")) << (int)sec_hand;
                  if (dbg) Serial << strcpy_P(p_buf, PSTR(", sec hands: ")) << (int)sec_hands;
                  if (dbg) Serial << endl;

                  Hands::instance()->setSecondHand(sec_hand, sec_hands);
                  if ( min_hand != Hands::instance()->minutes() ) 
                  {
                     refreshMinute();
                     Hands::instance()->setMinuteHand(min_hand);
                  }
                  else 
                  {
                     Hands::instance()->setMinuteHand(min_hand);
                  }
                  Hands::instance()->setHourHand(hrs_hand);
               }
            }
         }
      }

      /***********************************************/
      /*   The  jumper  to  adjust  the hands is in  */
      /*   place,  let them cycle through the hands  */
      /*   and set their lengths.                    */
      /***********************************************/
      else 
      {
         static uint8_t hand = 0;
         
         int8_t ch = switches.check();
         if ( ch >= 0 ) 
         {
            Serial << strcpy_P(p_buf, PSTR("ch: ")) << (int)ch << endl;
         }
   //      Serial << ".";
         switch(clock_state)
         {
            Serial << strcpy_P(p_buf, PSTR("clock_state: ")) << (int)clock_state << endl;
            case CLOCK_FORWARD:
            {
               if (dbg) Serial << strcpy_P(p_buf, PSTR("CLOCK_FORWARD: hand = ")) << (int)hand << endl;
               if ( ch == SW1_SHORT || SW1_ON ) 
               {
                  Serial << strcpy_P(p_buf, PSTR("CLOCK_FORWARD: hand = ")) << (int)hand << endl;
                  Hands::instance()->writeHand(hand, 0x00);
                  hand++;
                  if ( hand > 11 ) 
                  {
                     hand = 0;
                  }
               }
               else if ( ch == SW2_ON )
               {
                  clock_state = CLOCK_RUNNING;
                  eeprom.write();
               }
               else if ( ch == SW3_ON )
               {
                  Hands::instance()->handAdjust(hand, +3);
                  Serial << strcpy_P(p_buf, PSTR("++ hand: ")) << hand;
                  Serial << strcpy_P(p_buf, PSTR(", handAdjust = ")) << (int)settings->handAdjust[hand] << endl;
               }
               else if ( ch == SW4_ON )
               {
                  Hands::instance()->handAdjust(hand, -3);
                  Serial << strcpy_P(p_buf, PSTR("-- hand: ")) << hand;
                  Serial << strcpy_P(p_buf, PSTR(", handAdjust = ")) << (int)settings->handAdjust[hand] << endl;
               }
               Hands::instance()->writeHand(hand, 100);
            }
            break;

         }
         if ( clock_state == CLOCK_RUNNING ) 
         {
            if ( ch == SW1_ON || ch == SW1_OFF) 
            {
               Serial << strcpy_P(p_buf, PSTR("CLOCK_RUNNING: CLOCK_SW[12]_ON")) << endl;
               clock_state = CLOCK_FORWARD;
               Serial << strcpy_P(p_buf, PSTR("CLOCK_RUNNING: Moved to CLOCK_FORWARD")) << endl;
            }
            else if ( currentDate.getInterval(Clock::ONE_SECOND_TICK) > 2 )
            {
               Serial << strcpy_P(p_buf, PSTR("CLOCK_RUNNING: Using IN13 Lamps"));
               currentDate.resetInterval(Clock::ONE_SECOND_TICK);
               Hands::instance()->writeHand(hand, 0x00);
               hand++;
               if ( hand > 11 ) 
               {
                  hand = 0;
               }
               if (dbg) Serial << strcpy_P(p_buf, PSTR("hand = ")) << (int)hand << endl;
               Hands::instance()->writeHand(hand, 100);
            }
         }
      }
      gen_60hz.off();
   }
   return(0);
}
