/*
  switch.cpp - Kitschy Nixie Neon Clock

  Copyright (c) 2017 Joe Croft <croftj@gmail.com>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA

  $Id: //depot/kitchenclock/code/switch.cpp#5 $
*/

# include "switch.h"

# include <stdio.h>

# include "IOPort.h"

# include "core/HardwareSerial.h"

extern char p_buf[128];

InputPort switch_1;
InputPort switch_2;
InputPort switch_3;
InputPort switch_4;
OutputPort   debug;

uint16_t Switch::SW_ON_MIN;
uint16_t Switch::SW_ON_TRIGGER;
uint16_t Switch::SW_ON_MAX;
uint16_t Switch::SW_SHORT_MIN;
uint16_t Switch::SW_SHORT_MAX;
uint16_t Switch::SW_LONG_MIN;
uint16_t Switch::SW_LONG_MAX;

extern char ibuf[];

/**********************************************************************/
/*   Only 8 button presses will be saved at a time                    */
/**********************************************************************/
#define  SW_BUFFER_SIZE 8

struct ring_buffer {
  uint8_t buffer[SW_BUFFER_SIZE];
  int head;
  int tail;
};

/**********************************************************************/
/*   This is the actual queue for the switch presses                  */
/**********************************************************************/
ring_buffer swBuf; // = { { 0 }, 0, 0 };

/**********************************************************************/
/*   This  fnction  stores  the  passed in character into the switch  */
/*   queue.                                                           */
/**********************************************************************/
static void store_char(uint8_t c)
{
	int i = (swBuf.head + 1) % SW_BUFFER_SIZE;

   /***********************************************/
   /*   If  we are going to overwrite characters  */
   /*   already  in  the  queue, discard the new  */
   /*   presses.                                  */
   /***********************************************/
	if (i != swBuf.tail) {
//      Serial << "store_char(" << (int)c << ")" <<endl;
		swBuf.buffer[swBuf.head] = c;
		swBuf.head = i;
	}
}

/**********************************************************************/
/*   This  function pulls a character from the switch queue if is it  */
/*   present  and  returns  it  or  it returns -1 if no character is  */
/*   present.                                                         */
/**********************************************************************/
static uint8_t get_char(void)
{
	// if the head isn't ahead of the tail, we don't have any characters
	if (swBuf.head == swBuf.tail) {
		return -1;
	} else {
		unsigned char c = swBuf.buffer[swBuf.tail];
		swBuf.tail = (swBuf.tail + 1) % SW_BUFFER_SIZE;
		return c;
	}
}

/**********************************************************************/
/*   This is the constructor of the Switch object.                    */
/**********************************************************************/
Switch::Switch() :
      m_enabled(true)
{
   /***********************************************/
   /*   Set up out psuedo constants               */
   /***********************************************/
   SW_ON_MIN      = 4;
   SW_ON_TRIGGER  = 15;
   SW_ON_MAX      = 32;
   SW_SHORT_MIN   = 62;
   SW_SHORT_MAX   = 160;
   SW_LONG_MIN    = 200;
   SW_LONG_MAX    = 1010;

   m_swCount[0] = m_swTime[0] = 0;
   m_swCount[1] = m_swTime[1] = 0;
   m_swCount[2] = m_swTime[2] = 0;
   m_swCount[3] = m_swTime[3] = 0;
   memset (&swBuf, '\0', sizeof(swBuf));

   switch_1.SetPort(IOPort::PORT_C, _BV(1));
   switch_1.on();
   m_swState[0] = switch_1.read();

   switch_2.SetPort(IOPort::PORT_D, _BV(7));
   switch_2.on();
   m_swState[1] = switch_2.read();

   switch_3.SetPort(IOPort::PORT_D, _BV(3));
   switch_3.on();
   m_swState[2] = switch_3.read();

   switch_4.SetPort(IOPort::PORT_C, _BV(2));
   switch_4.on();
   m_swState[3] = switch_4.read();

   debug.SetPort(IOPort::PORT_C, _BV(3));
   debug.off();
}

/**********************************************************************/
/*   This   function   eith   enables  of  disablesthe  auto  repeat  */
/*   functionality.                                                   */
/**********************************************************************/
void Switch::setAutoRepeat(uint8_t f)
{
   for (uint8_t x = 0; x < SWITCH_COUNT; x++)
   {
      m_autoRepeat[x] = f;
   }
}

void Switch::setAutoRepeat(uint8_t sw, uint8_t f)
{
   m_autoRepeat[sw] = f;
}

/**********************************************************************/
/*   This function looks at the pins for the switches and determines  */
/*   which  switches  have  been  pressed  and  puts the appropriate  */
/*   character codes for any pressed switches into the queue.         */
/**********************************************************************/
void Switch::update()
{
//   if ( m_enabled == true )
   {
      debug.on();
      uint8_t btns[4];
      btns[0] = switch_1.read();
      btns[1] = switch_2.read();
   //   btns[2] = 0;
      btns[2] = switch_3.read();
   //   btns[3] = 0;
      btns[3] = switch_4.read();

      for (uint8_t sw = 0; sw < SWITCH_COUNT; sw++)
      {
         if ( btns[sw] ^ m_swState[sw] )
         {
            m_swTime[sw] = 0;
            if ( m_swCount[sw] > 2 )
            {
               m_swState[sw] = btns[sw];
            }
            else
            {
               m_swCount[sw]++;
            }
         }

         /***********************************************/
         /*   Otherwise,  if  the  switch  is pressed,  */
         /*   increment the length of time it has been  */
         /*   pressed,  make  sure the counter doesn't  */
         /*   roll over                                 */
         /***********************************************/
         else if (m_swTime[sw] < 60000)
         {
            m_swCount[sw] = 0;
            m_swTime[sw]++;
         }

         /***********************************************/
         /*   Now  see  if the current switch has been  */
         /*   on  for better than the minimum time but  */
         /*   less  than  the maximum time. If so, put  */
         /*   it's  count one past so that it won't be  */
         /*   counted  as  a switch on event again and  */
         /*   push  a  switch  on  character  for  the  */
         /*   switch into the queue.                    */
         /***********************************************/
         if ( btns[sw] == 0 )
         {
            if ( (m_swTime[sw] > SW_ON_MIN && m_swTime[sw] < SW_ON_TRIGGER) )
            {
                  Serial << strcpy_P(p_buf, PSTR(" SW_ON swTime[")) << (int)sw << "] = " << (int)m_swTime[sw] << endl;
                  m_swTime[sw] = SW_ON_TRIGGER + 1;
                  store_char(sw * SWITCH_STATES + SW1_ON);
            }

            /***********************************************/
            /*   If we are autorepeat and we are past the  */
            /*   trigger point, Send a SWn_SHORT. In this  */
            /*   way  we  can  still  detect when someone  */
            /*   presses  two  buttons  at once but still  */
            /*   have a timely response from 'triggering'  */
            /*   some action with SWn_SHORT.               */
            /***********************************************/
            else if ( (m_swTime[sw] > SW_ON_TRIGGER && m_swTime[sw] < SW_ON_MAX) )
            {
               if ( m_autoRepeat[sw] )
               {
                  Serial << strcpy_P(p_buf, PSTR(" SW_SHORT swTime[")) << (int)sw << "] = " << (int)m_swTime[sw] << endl;
                  store_char(sw * SWITCH_STATES + SW1_SHORT);
                  m_swTime[sw] = SW_ON_MAX + 1;
               }
            }
            /***********************************************/
            /*   If  we  went  past the time for a short   */
            /*   press  but  not  longer  than its ranges  */
            /*   max,  show  the button was pressed for a  */
            /*   short time.                               */
            /***********************************************/
            else if (m_swTime[sw] > SW_SHORT_MIN && m_swTime[sw] < SW_SHORT_MAX )
            {
               store_char(sw * SWITCH_STATES + SW1_SHORT);

               /***********************************************/
               /*   If auto repeat is enabled, rest the time  */
               /*   back  to  the before start of this range  */
               /*   so we will end up here again              */
               /***********************************************/
               if ( m_autoRepeat[sw] )
                  m_swTime[sw] = SW_ON_MAX + 1;

               /***********************************************/
               /*   Otherwise,   move  past  this  range  to  */
               /*   capture long button presses.              */
               /***********************************************/
               else
                  m_swTime[sw] = SW_SHORT_MAX + 1;
               Serial << strcpy_P(p_buf, PSTR("m_autoRepeat[")) << (int)sw << "] = " << (int)m_autoRepeat[sw];
               Serial << strcpy_P(p_buf, PSTR(", SW_SHORT swTime[")) << (int)sw << "] = " << (int)m_swTime[sw] << endl;
            }

            /***********************************************/
            /*   If  button  has been pressed a good long  */
            /*   time,  show  we  have  had a long button  */
            /*   press  and  make sure we set the time so  */
            /*   we only give one press.                   */
            /***********************************************/
            else if (m_swTime[sw] > SW_LONG_MIN && m_swTime[sw] < SW_LONG_MAX )
            {
               m_swTime[sw] = SW_LONG_MAX + 1;
               store_char(sw * SWITCH_STATES + SW1_LONG);
               Serial << strcpy_P(p_buf, PSTR(" SW_LONG swTime[")) << (int)sw << "] = " << (int)m_swTime[sw] << endl;
            }
         }

         /***********************************************/
         /*   Finaly,  if the button has been released  */
         /*   and  the  time  it was on for was in the  */
         /*   range  for  a regular button press, show  */
         /*   that the button was released.             */
         /***********************************************/
         else if (  btns[sw] != 0 && m_swTime[sw] > SW_ON_MIN && m_swTime[sw] < SW_ON_MAX )
         {
            m_swTime[sw] = SW_ON_MAX + 1;
            store_char(sw * SWITCH_STATES + SW1_OFF);
            Serial << strcpy_P(p_buf, PSTR(" SW_OFF swTime[")) << (int)sw << "] = " << (int)m_swTime[sw] << endl;
         }
      }
      debug.off();
   }
}


/**********************************************************************/
/*   THis  function  simply returns the next switch character in the  */
/*   queue if any is present.                                         */
/**********************************************************************/
uint8_t Switch::check()
{
   return(get_char());
}

void Switch::printTimes()
{
   for (uint8_t sw = 0; sw < SWITCH_COUNT; sw++)
   {
      Serial << strcpy_P(p_buf, PSTR("swTime[")) << (int)sw << "] = " << (uint32_t)m_swTime[sw] << endl;
   }
}

Switch switches;
